#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/cacheflush.h>


#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ide.h>

#include "Types.h"

#include "agentlite.h"
#include "support_fpga.h"
#include "support_fpga_hw.h"
#include "support_fpga_defs.h"

#if 0
#define ADC_MEM_CHECK_REG_FAILS		0xBB
#define ADC_REQ_MEM_REG_FAILS		0xCC
#define ADC_MEM_REMAP_FAILS			0x0A
#define ADC_REG_CHR_DEV_FAILS 		0xAA
#define SET_ADC_OPTION				4
#endif

#define MAX_RETRIES 3
#define INVALID_EXCHANGE_BUS_WAIT_PERIOD  HZ/(10 * MAX_RETRIES)



#define SUPPORT_FPGA_MEM_REMAP_FAILS 0xAB 
#define SUPPORT_FPGA_MEM_CHECK_REG_FAILS 0xAC 
#define SUPPORT_FPGA_REQ_MEM_REG_FAILS 0xAD 
#define SUPPORT_FPGA_REG_CHR_DEV_FAILS 		0xAE





#ifdef SIMULATION
static unsigned short simu_data[256] = {0};
#endif



/* ------------------------------------------------------------------------- *
 *  Function prototypes
 * ------------------------------------------------------------------------- */



static int support_fpga_init(void);
static void support_fpga_exit(void );
static int support_fpga_module_open(struct inode *inode, struct file *file);
static int support_fpga_module_close(struct inode *inode, struct file *file);
static long support_fpga_module_ioctl(struct file *file, uint cmd, ulong arg);
extern int write_init_st(int option,int value);



/* -------------------------------------------------------------------------
 *  Static variables
 * ------------------------------------------------------------------------- */
static struct file_operations support_fpga_module_ops;
static void __iomem *support_fpga_parallel_bus_remap_addr;
static short init_support_fpga_status=0;



/* SUPPORT FPGA Driver Private data */
typedef struct {
	int (*read_parallel_bus_reg)  (uint32_t *support_fpga_value, uint32_t reg_offset);
	int (*write_parallel_bus_reg) (uint32_t support_fpga_value, uint32_t reg_offset);

} support_fpga_drv_t;




static int
support_fpga_init(void)
{
	int err,ret=0;
	uint32_t support_fpga_base;


	memset(&support_fpga_module_ops,0,sizeof(support_fpga_module_ops));
	support_fpga_module_ops.open=support_fpga_module_open;
	support_fpga_module_ops.release=support_fpga_module_close;
	support_fpga_module_ops.unlocked_ioctl= support_fpga_module_ioctl;
	support_fpga_module_ops.read=NULL;
	support_fpga_module_ops.write=NULL;
	support_fpga_module_ops.owner=THIS_MODULE;

    /* ----  Register the character device ------------------- */
	if ((err = register_chrdev(SUPPORT_FPGA_MAJOR_NUM,DRIVER_NAME,&support_fpga_module_ops)) < 0) {
		printk (KERN_ERR "support_fpga.c: SUPPORT FPGA register_chrdrv failed. err = %d\n", err);
		init_support_fpga_status=SUPPORT_FPGA_REG_CHR_DEV_FAILS;
		return -ENODEV;
	}

	/* Remap registers */
	 support_fpga_base= SUPPORT_FPGA_BASE_OFFSET;
	if ((support_fpga_parallel_bus_remap_addr = ioremap(support_fpga_base, SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE)) == NULL) {
		printk(KERN_ERR "failed to map SUPPORT FPGA  IO space %08x-%08x to memory",support_fpga_base, support_fpga_base + SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE - 1);
		init_support_fpga_status=SUPPORT_FPGA_MEM_REMAP_FAILS;
		ret = -EIO;
		support_fpga_exit();
		return ret;
	}

	/* check Support FPGA memory region */
	if ((err = check_mem_region(support_fpga_base, SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE)) < 0)  {
		printk(KERN_ERR "Support FPGA IO space %08x-%08x already in use (err %d)",support_fpga_base, support_fpga_base + SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE - 1, err);
		init_support_fpga_status=SUPPORT_FPGA_MEM_CHECK_REG_FAILS;
		ret = -EBUSY;
		support_fpga_exit();
		return ret;
	}

	/* request for the memory region */	
	if (request_mem_region(support_fpga_base,SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE,DRIVER_NAME)  == NULL) { 
		printk(KERN_ERR "IO space %08x-%08x adc request mem region failed (err %d)",support_fpga_base, support_fpga_base + SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE - 1, err);
		init_support_fpga_status=SUPPORT_FPGA_REQ_MEM_REG_FAILS;
		ret = -ENODEV;
		support_fpga_exit();
		return ret;
	}



	//write_init_st(SET_ADC_OPTION,init_adc_status);
	printk(KERN_ALERT "Support FPGA Module was  sucessully loaded %px\n", (unsigned long *)&support_fpga_parallel_bus_remap_addr);
	return ret;
}


void
support_fpga_exit(void ) {
	uint32_t support_fpga_base;

	printk (KERN_ALERT "Unloading Support FPGA driver ..\n");
	support_fpga_base = SUPPORT_FPGA_BASE_OFFSET;

	if(init_support_fpga_status == 0) {
		init_support_fpga_status = 0xff;
	}

	/* release the memory occupied */
	release_mem_region(support_fpga_base,SUPPORT_FPGA_PARALLEL_BUS_REG_SIZE);

	/* Unmap registers */
	iounmap(support_fpga_parallel_bus_remap_addr);
	support_fpga_parallel_bus_remap_addr = NULL;

	/* ----  UnRegister the character device ------------------- */
	unregister_chrdev(SUPPORT_FPGA_MAJOR_NUM,DRIVER_NAME);

	return ;

}



int 
support_fpga_module_open(struct inode *inode, struct file *file) {

   support_fpga_drv_t *support_fpga_drv  = (support_fpga_drv_t*) kmalloc ( sizeof(support_fpga_drv_t), GFP_KERNEL );
	memset(support_fpga_drv, 0, sizeof(support_fpga_drv_t));


	/* Initialize the private data */	
	support_fpga_drv->read_parallel_bus_reg = support_fpga_read_parallel_bus_reg;
	support_fpga_drv->write_parallel_bus_reg = support_fpga_write_parallel_bus_reg;

	file->private_data = support_fpga_drv;



	return 0;
}


int 
support_fpga_module_close(struct inode *inode, struct file *file) {
	support_fpga_drv_t *support_fpga_drv;
	support_fpga_drv = file->private_data;
	file->private_data = NULL;
	kfree(support_fpga_drv);


	return 0;
}


long 
support_fpga_module_ioctl(struct file *file, uint cmd, ulong arg) {
	support_fpga_drv_t *support_fpga_drv;
	reg_value_lml_t  reg_value;
 

	support_fpga_drv =file->private_data; 

	switch (cmd) {	
		case READ_REG :
			copy_from_user ((void*)&reg_value, (void *)arg,sizeof (reg_value_lml_t));
	#ifdef SIMULATION
				reg_value.reg_offset_addr = reg_value.reg_offset_addr & 0xFF; // simulation data only has 256 bytes.
				reg_value.reg_value = simu_data [reg_value.reg_offset_addr];

	#else 
				if (support_fpga_drv->read_parallel_bus_reg)
				{
					if (0 != support_fpga_drv->read_parallel_bus_reg(&reg_value.reg_value,reg_value.reg_offset_addr))
					return -1;
				}
			
	#endif
			copy_to_user ((void*)arg, (void*)&reg_value, sizeof (reg_value_lml_t));
			break;

		case WRITE_REG :
			copy_from_user ((void*)&reg_value, (void *)arg,sizeof (reg_value_lml_t));
	#ifdef SIMULATION
			reg_value.reg_offset_addr = reg_value.reg_offset_addr & 0xFF; // simulation data only has 256 bytes.
			simu_data [reg_value.reg_offset_addr] = reg_value.reg_value;
	#else 

			if (support_fpga_drv->write_parallel_bus_reg)
			{
				if (0 != support_fpga_drv->write_parallel_bus_reg(reg_value.reg_value,reg_value.reg_offset_addr))
				return -1;
			}
	#endif
			break;

		default :
			printk (KERN_ALERT "Invalid ADC driver ioctl\n");
			return -EINVAL;
		}

	return 0;
}


module_init(support_fpga_init);
module_exit(support_fpga_exit);

int
support_fpga_read_parallel_bus_reg(uint32_t *support_fpga_value, uint32_t reg_off_addr) {
	void __iomem *remapped_addr ;

	remapped_addr = support_fpga_parallel_bus_remap_addr + reg_off_addr;
	*support_fpga_value = *((unsigned short *)(remapped_addr));

	return 0;
}

int
support_fpga_write_parallel_bus_reg(uint32_t support_fpga_value, uint32_t reg_off_addr) {
	*((uint16_t *)(support_fpga_parallel_bus_remap_addr + reg_off_addr)) = (uint16_t) support_fpga_value;
	return 0;
}

EXPORT_SYMBOL_GPL(support_fpga_read_parallel_bus_reg);
EXPORT_SYMBOL_GPL(support_fpga_write_parallel_bus_reg);



MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Support FPGA Driver");
